from hccinfhir.datamodels import Demographics, ModelName
from typing import Optional

def has_any_hcc(hcc_list: list[str], hcc_set: set[str]) -> int:
    """Returns 1 if any HCC in the list is present, 0 otherwise"""
    return int(bool(set(hcc_list) & hcc_set))

def create_demographic_interactions(demographics: Demographics) -> dict:
    """Creates common demographic-based interactions"""
    interactions = {}
    is_female = demographics.category.startswith('F')
    is_male = demographics.category.startswith('M')
    is_aged = not demographics.non_aged
    
    # Original Disability interactions
    if is_aged:
        interactions['OriginallyDisabled_Female'] = int(demographics.orig_disabled) * int(is_female)
        interactions['OriginallyDisabled_Male'] = int(demographics.orig_disabled) * int(is_male)
    else:
        interactions['OriginallyDisabled_Female'] = 0
        interactions['OriginallyDisabled_Male'] = 0

    # LTI interactions - used for ESRD models
    if demographics.lti:
        interactions['LTI_Aged'] = int(is_aged)
        interactions['LTI_NonAged'] = int(not is_aged)
    else:
        interactions['LTI_Aged'] = 0
        interactions['LTI_NonAged'] = 0
        
    nemcaid = False
    if demographics.new_enrollee and demographics.dual_elgbl_cd in {'01', '02', '03', '04', '05', '06', '08'}:
        nemcaid = True
    ne_origds = int(demographics.age >= 65 and (demographics.orec is not None and demographics.orec == "1"))
    
    # Four mutually exclusive groups
    interactions.update({
        f'NMCAID_NORIGDIS_{demographics.category}': int(not nemcaid and not ne_origds),
        f'MCAID_NORIGDIS_{demographics.category}': int(nemcaid and not ne_origds),
        f'NMCAID_ORIGDIS_{demographics.category}': int(not nemcaid and ne_origds),
        f'MCAID_ORIGDIS_{demographics.category}': int(nemcaid and ne_origds)
    })

    return interactions

def create_dual_interactions(demographics: Demographics) -> dict:
    """Creates dual status interactions"""
    interactions = {}
    is_female = demographics.category.startswith('F')
    is_male = demographics.category.startswith('M')
    is_aged = not demographics.non_aged
    
    if demographics.fbd:
        interactions.update({
            'FBDual_Female_Aged': int(is_female) * int(is_aged),
            'FBDual_Female_NonAged': int(is_female) * int(not is_aged),
            'FBDual_Male_Aged': int(is_male) * int(is_aged),
            'FBDual_Male_NonAged': int(is_male) * int(not is_aged)
        })
    
    if demographics.pbd:
        interactions.update({
            'PBDual_Female_Aged': int(is_female) * int(is_aged),
            'PBDual_Female_NonAged': int(is_female) * int(not is_aged),
            'PBDual_Male_Aged': int(is_male) * int(is_aged),
            'PBDual_Male_NonAged': int(is_male) * int(not is_aged)
        })
        
    return interactions

def create_hcc_counts(hcc_set: set[str]) -> dict:
    """Creates HCC count variables"""
    counts = {}
    hcc_count = len(hcc_set)
    
    for i in range(1, 10):
        counts[f'D{i}'] = int(hcc_count == i)
    counts['D10P'] = int(hcc_count >= 10)
    
    return counts

def get_diagnostic_categories(model_name: ModelName, hcc_set: set[str]) -> dict:
    """Creates disease categories based on model version"""
    if model_name == "CMS-HCC Model V28":
        return {
            'CANCER_V28': has_any_hcc(['17', '18', '19', '20', '21', '22', '23'], hcc_set),
            'DIABETES_V28': has_any_hcc(['35', '36', '37', '38'], hcc_set),
            'CARD_RESP_FAIL_V28': has_any_hcc(['211', '212', '213'], hcc_set),
            'HF_V28': has_any_hcc(['221', '222', '223', '224', '225', '226'], hcc_set),
            'CHR_LUNG_V28': has_any_hcc(['276', '277', '278', '279', '280'], hcc_set),
            'KIDNEY_V28': has_any_hcc(['326', '327', '328', '329'], hcc_set),
            'SEPSIS_V28': int('2' in hcc_set),
            'gSubUseDisorder_V28': has_any_hcc(['135', '136', '137', '138', '139'], hcc_set),
            'gPsychiatric_V28': has_any_hcc(['151', '152', '153', '154', '155'], hcc_set),
            'NEURO_V28': has_any_hcc(['180', '181', '182', '190', '191', '192', '195', '196', '198', '199'], hcc_set),
            'ULCER_V28': has_any_hcc(['379', '380', '381', '382'], hcc_set)
        }
    elif model_name == "CMS-HCC Model V24":
        return {
            'CANCER': has_any_hcc(['8', '9', '10', '11', '12'], hcc_set),
            'DIABETES': has_any_hcc(['17', '18', '19'], hcc_set),
            'CARD_RESP_FAIL': has_any_hcc(['82', '83', '84'], hcc_set),
            'CHF': int('85' in hcc_set),
            'gCopdCF': has_any_hcc(['110', '111', '112'], hcc_set),
            'RENAL_V24': has_any_hcc(['134', '135', '136', '137', '138'], hcc_set), 
            'SEPSIS': int('2' in hcc_set),
            'gSubstanceUseDisorder_V24': has_any_hcc(['54', '55', '56'], hcc_set),
            'gPsychiatric_V24': has_any_hcc(['57', '58', '59', '60'], hcc_set),
            'PRESSURE_ULCER': has_any_hcc(['157', '158', '159'],  hcc_set) # added in 2018-11-20
        }
    elif model_name == "CMS-HCC Model V22":
        return {
            'CANCER': has_any_hcc(['8', '9', '10', '11', '12'], hcc_set),
            'DIABETES': has_any_hcc(['17', '18', '19'], hcc_set),
            'CARD_RESP_FAIL': has_any_hcc(['82', '83', '84'], hcc_set),
            'CHF': int('85' in hcc_set),
            'gCopdCF': has_any_hcc(['110', '111', '112'], hcc_set),
            'RENAL': has_any_hcc(['134', '135', '136', '137'], hcc_set), 
            'SEPSIS': int('2' in hcc_set),
            'gSubstanceUseDisorder': has_any_hcc(['54', '55'], hcc_set),
            'gPsychiatric': has_any_hcc(['57', '58'], hcc_set),
            'PRESSURE_ULCER': has_any_hcc(['157', '158'],  hcc_set) # added in 2012-10-19
        }
    elif model_name == "CMS-HCC ESRD Model V24":
        return {
            'CANCER': has_any_hcc(['8', '9', '10', '11', '12'], hcc_set),
            'DIABETES': has_any_hcc(['17', '18', '19'], hcc_set),
            'CARD_RESP_FAIL': has_any_hcc(['82', '83', '84'], hcc_set),
            'CHF': int('85' in hcc_set),            
            'gCopdCF': has_any_hcc(['110', '111', '112'], hcc_set),
            'RENAL_V24': has_any_hcc(['134', '135', '136', '137', '138'], hcc_set),
            'SEPSIS': int('2' in hcc_set),
            'PRESSURE_ULCER': has_any_hcc(['157', '158', '159', '160'], hcc_set), # added in 2018-11-20
            'gSubstanceUseDisorder_V24': has_any_hcc(['54', '55', '56'], hcc_set),
            'gPsychiatric_V24': has_any_hcc(['57', '58', '59', '60'], hcc_set)
        }
    elif model_name == "CMS-HCC ESRD Model V21":
        return {
            'CANCER': has_any_hcc(['8', '9', '10', '11', '12'], hcc_set),
            'DIABETES': has_any_hcc(['17', '18', '19'], hcc_set),
            'IMMUNE': int('47' in hcc_set),
            'CARD_RESP_FAIL': has_any_hcc(['82', '83', '84'], hcc_set),
            'CHF': int('85' in hcc_set),
            'COPD': has_any_hcc(['110', '111'], hcc_set),
            'RENAL': has_any_hcc(['134', '135', '136', '137', '138', '139', '140', '141'], hcc_set),
            'COMPL': int('176' in hcc_set),
            'SEPSIS': int('2' in hcc_set), 
            'PRESSURE_ULCER': has_any_hcc(['157', '158', '159', '160'], hcc_set)
        }
    elif model_name == "RxHCC Model V08":
        # RxModel doesn't seem to have any diagnostic category interactions
        return {}
    return {}

def create_disease_interactions(model_name: ModelName, 
                              diagnostic_cats: dict, 
                              demographics: Optional[Demographics],
                              hcc_set: Optional[set[str]]) -> dict:
    """Creates disease interaction variables based on model version.
    
    Args:
        model_name: The HCC model version being used
        diagnostic_cats: Dictionary of diagnostic categories
        demographics: Optional demographic information for age/sex/disability interactions
        hcc_set: Optional set of HCCs for direct HCC checks
        
    Returns:
        Dictionary containing all disease interaction variables
    """
    interactions = {}
    
    if model_name == "CMS-HCC Model V28":
        # Base V28 disease interactions
        interactions.update({
            'DIABETES_HF_V28': diagnostic_cats['DIABETES_V28'] * diagnostic_cats['HF_V28'],
            'HF_CHR_LUNG_V28': diagnostic_cats['HF_V28'] * diagnostic_cats['CHR_LUNG_V28'],
            'HF_KIDNEY_V28': diagnostic_cats['HF_V28'] * diagnostic_cats['KIDNEY_V28'],
            'CHR_LUNG_CARD_RESP_FAIL_V28': diagnostic_cats['CHR_LUNG_V28'] * diagnostic_cats['CARD_RESP_FAIL_V28'],
            'gSubUseDisorder_gPsych_V28': diagnostic_cats['gSubUseDisorder_V28'] * diagnostic_cats['gPsychiatric_V28'],
            'DISABLED_CANCER_V28': demographics.disabled * diagnostic_cats['CANCER_V28'],
            'DISABLED_NEURO_V28': demographics.disabled * diagnostic_cats['NEURO_V28'],
            'DISABLED_HF_V28': demographics.disabled * diagnostic_cats['HF_V28'],
            'DISABLED_CHR_LUNG_V28': demographics.disabled * diagnostic_cats['CHR_LUNG_V28'],
            'DISABLED_ULCER_V28': demographics.disabled * diagnostic_cats['ULCER_V28']
        })
            
    elif model_name == "CMS-HCC Model V24":
        # Base V24/V22 disease interactions
        interactions.update({
            'HCC47_gCancer': int('47' in hcc_set) * diagnostic_cats['CANCER'],
            'DIABETES_CHF': diagnostic_cats['DIABETES'] * diagnostic_cats['CHF'],
            'CHF_gCopdCF': diagnostic_cats['CHF'] * diagnostic_cats['gCopdCF'],
            'HCC85_gRenal_V24': diagnostic_cats['CHF'] * diagnostic_cats['RENAL_V24'],
            'gCopdCF_CARD_RESP_FAIL': diagnostic_cats['gCopdCF'] * diagnostic_cats['CARD_RESP_FAIL'],
            'HCC85_HCC96': int('85' in hcc_set) * int('96' in hcc_set),
            'gSubstanceAbuse_gPsych': diagnostic_cats['gSubstanceUseDisorder_V24'] * diagnostic_cats['gPsychiatric_V24'],
            'SEPSIS_PRESSURE_ULCER': diagnostic_cats['SEPSIS'] * diagnostic_cats['PRESSURE_ULCER'],
            'SEPSIS_ARTIF_OPENINGS': diagnostic_cats['SEPSIS'] * int('188' in hcc_set),
            'ART_OPENINGS_PRESS_ULCER': int('188' in hcc_set) * diagnostic_cats['PRESSURE_ULCER'],
            'gCopdCF_ASP_SPEC_B_PNEUM': diagnostic_cats['gCopdCF'] * int('114' in hcc_set),
            'ASP_SPEC_B_PNEUM_PRES_ULC': int('114' in hcc_set) * diagnostic_cats['PRESSURE_ULCER'],
            'SEPSIS_ASP_SPEC_BACT_PNEUM': diagnostic_cats['SEPSIS'] * int('114' in hcc_set),
            'SCHIZOPHRENIA_gCopdCF': int('57' in hcc_set) * diagnostic_cats['gCopdCF'],
            'SCHIZOPHRENIA_CHF': int('57' in hcc_set) * diagnostic_cats['CHF'],
            'SCHIZOPHRENIA_SEIZURES': int('57' in hcc_set) * int('79' in hcc_set),
            'DISABLED_HCC85': demographics.disabled * int('85' in hcc_set),
            'DISABLED_PRESSURE_ULCER': demographics.disabled * diagnostic_cats['PRESSURE_ULCER'],
            'DISABLED_HCC161': demographics.disabled * int('161' in hcc_set),
            'DISABLED_HCC39': demographics.disabled * int('39' in hcc_set),
            'DISABLED_HCC77': demographics.disabled * int('77' in hcc_set),
            'DISABLED_HCC6': demographics.disabled * int('6' in hcc_set)
        })
    elif model_name == "CMS-HCC Model V22":
        # Base V24/V22 disease interactions
        interactions.update({
            'HCC47_gCancer': int('47' in hcc_set) * diagnostic_cats['CANCER'],
            'HCC85_gDiabetesMellitus': int('85' in hcc_set) * diagnostic_cats['DIABETES'],
            'HCC85_gCopdCF': int('85' in hcc_set) * diagnostic_cats['gCopdCF'],
            'HCC85_gRenal': int('85' in hcc_set) * diagnostic_cats['RENAL'],
            'gRespDepandArre_gCopdCF': diagnostic_cats['CARD_RESP_FAIL'] * diagnostic_cats['gCopdCF'],
            'HCC85_HCC96': int('85' in hcc_set) * int('96' in hcc_set),
            'gSubstanceAbuse_gPsychiatric': diagnostic_cats['gSubstanceUseDisorder'] * diagnostic_cats['gPsychiatric'],
            'DIABETES_CHF': diagnostic_cats['DIABETES'] * diagnostic_cats['CHF'],
            'CHF_gCopdCF': diagnostic_cats['CHF'] * diagnostic_cats['gCopdCF'],
            'gCopdCF_CARD_RESP_FAIL': diagnostic_cats['gCopdCF'] * diagnostic_cats['CARD_RESP_FAIL'],
            'SEPSIS_PRESSURE_ULCER': diagnostic_cats['SEPSIS'] * diagnostic_cats['PRESSURE_ULCER'],
            'SEPSIS_ARTIF_OPENINGS': diagnostic_cats['SEPSIS'] * int('188' in hcc_set),
            'ART_OPENINGS_PRESSURE_ULCER': int('188' in hcc_set) * diagnostic_cats['PRESSURE_ULCER'],
            'DIABETES_CHF': diagnostic_cats['DIABETES'] * diagnostic_cats['CHF'],
            'gCopdCF_ASP_SPEC_BACT_PNEUM': diagnostic_cats['gCopdCF'] * int('114' in hcc_set),
            'ASP_SPEC_BACT_PNEUM_PRES_ULC': int('114' in hcc_set) * diagnostic_cats['PRESSURE_ULCER'],
            'SEPSIS_ASP_SPEC_BACT_PNEUM': diagnostic_cats['SEPSIS'] * int('114' in hcc_set),
            'SCHIZOPHRENIA_gCopdCF': int('57' in hcc_set) * diagnostic_cats['gCopdCF'],
            'SCHIZOPHRENIA_CHF': int('57' in hcc_set) * diagnostic_cats['CHF'],
            'SCHIZOPHRENIA_SEIZURES': int('57' in hcc_set) * int('79' in hcc_set),
            'DISABLED_HCC85': demographics.disabled * int('85' in hcc_set),
            'DISABLED_PRESSURE_ULCER': demographics.disabled * diagnostic_cats['PRESSURE_ULCER'],
            'DISABLED_HCC161': demographics.disabled * int('161' in hcc_set),
            'DISABLED_HCC39': demographics.disabled * int('39' in hcc_set),
            'DISABLED_HCC77': demographics.disabled * int('77' in hcc_set),
            'DISABLED_HCC6': demographics.disabled * int('6' in hcc_set)
        })
    elif model_name == "CMS-HCC ESRD Model V24":
        # Base ESRD V24 disease interactions
        interactions.update({
            'HCC47_gCancer': int('47' in hcc_set) * diagnostic_cats['CANCER'],
            'DIABETES_CHF': diagnostic_cats['DIABETES'] * diagnostic_cats['CHF'],
            'CHF_gCopdCF': diagnostic_cats['CHF'] * diagnostic_cats['gCopdCF'],
            'HCC85_gRenal_V24': int('85' in hcc_set) * diagnostic_cats['RENAL_V24'],
            'gCopdCF_CARD_RESP_FAIL': diagnostic_cats['gCopdCF'] * diagnostic_cats['CARD_RESP_FAIL'],
            'HCC85_HCC96': int('85' in hcc_set) * int('96' in hcc_set),
            'gSubUseDs_gPsych_V24': diagnostic_cats['gSubstanceUseDisorder_V24'] * diagnostic_cats['gPsychiatric_V24'],
            'NONAGED_gSubUseDs_gPsych': demographics.non_aged * (diagnostic_cats['gSubstanceUseDisorder_V24'] * diagnostic_cats['gPsychiatric_V24']),
            'NONAGED_HCC6': demographics.non_aged * int('6' in hcc_set),
            'NONAGED_HCC34': demographics.non_aged * int('34' in hcc_set),
            'NONAGED_HCC46': demographics.non_aged * int('46' in hcc_set),
            'NONAGED_HCC110': demographics.non_aged * int('110' in hcc_set),
            'NONAGED_HCC176': demographics.non_aged * int('176' in hcc_set),
            'SEPSIS_PRESSURE_ULCER_V24': diagnostic_cats['SEPSIS'] * diagnostic_cats['PRESSURE_ULCER'],
            'SEPSIS_ARTIF_OPENINGS': diagnostic_cats['SEPSIS'] * int('188' in hcc_set),
            'ART_OPENINGS_PRESS_ULCER_V24': int('188' in hcc_set) * diagnostic_cats['PRESSURE_ULCER'],
            'gCopdCF_ASP_SPEC_B_PNEUM': diagnostic_cats['gCopdCF'] * int('114' in hcc_set),
            'ASP_SPEC_B_PNEUM_PRES_ULC_V24': int('114' in hcc_set) * diagnostic_cats['PRESSURE_ULCER'],
            'SEPSIS_ASP_SPEC_BACT_PNEUM': diagnostic_cats['SEPSIS'] * int('114' in hcc_set),
            'SCHIZOPHRENIA_gCopdCF': int('57' in hcc_set) * diagnostic_cats['gCopdCF'],
            'SCHIZOPHRENIA_CHF': int('57' in hcc_set) * diagnostic_cats['CHF'],
            'SCHIZOPHRENIA_SEIZURES': int('57' in hcc_set) * int('79' in hcc_set),
            'NONAGED_HCC85': demographics.non_aged * int('85' in hcc_set),
            'NONAGED_PRESSURE_ULCER_V24': demographics.non_aged * diagnostic_cats['PRESSURE_ULCER'],
            'NONAGED_HCC161': demographics.non_aged * int('161' in hcc_set),
            'NONAGED_HCC39': demographics.non_aged * int('39' in hcc_set),
            'NONAGED_HCC77': demographics.non_aged * int('77' in hcc_set)
        })
    
    elif model_name == 'CMS-HCC ESRD Model V21':
        # ESRD Community model interactions
        interactions.update({
            'SEPSIS_CARD_RESP_FAIL': diagnostic_cats['SEPSIS'] * diagnostic_cats['CARD_RESP_FAIL'],
            'CANCER_IMMUNE': diagnostic_cats['CANCER'] * diagnostic_cats['IMMUNE'],
            'DIABETES_CHF': diagnostic_cats['DIABETES'] * diagnostic_cats['CHF'],
            'CHF_COPD': diagnostic_cats['CHF'] * diagnostic_cats['COPD'],
            'CHF_RENAL': diagnostic_cats['CHF'] * diagnostic_cats['RENAL'],
            'COPD_CARD_RESP_FAIL': diagnostic_cats['COPD'] * diagnostic_cats['CARD_RESP_FAIL'],
            'NONAGED_HCC6': demographics.non_aged * int('6' in hcc_set),
            'NONAGED_HCC34': demographics.non_aged * int('34' in hcc_set),
            'NONAGED_HCC46': demographics.non_aged * int('46' in hcc_set),
            'NONAGED_HCC54': demographics.non_aged * int('54' in hcc_set),
            'NONAGED_HCC55': demographics.non_aged * int('55' in hcc_set),
            'NONAGED_HCC110': demographics.non_aged * int('110' in hcc_set),
            'NONAGED_HCC176': demographics.non_aged * int('176' in hcc_set),
            'SEPSIS_PRESSURE_ULCER': diagnostic_cats['SEPSIS'] * diagnostic_cats['PRESSURE_ULCER'],
            'SEPSIS_ARTIF_OPENINGS': diagnostic_cats['SEPSIS'] * int('188' in hcc_set),
            'ART_OPENINGS_PRESSURE_ULCER': int('188' in hcc_set) * diagnostic_cats['PRESSURE_ULCER'],
            'DIABETES_CHF': diagnostic_cats['DIABETES'] * diagnostic_cats['CHF'],
            'COPD_ASP_SPEC_BACT_PNEUM': diagnostic_cats['COPD'] * int('114' in hcc_set),
            'ASP_SPEC_BACT_PNEUM_PRES_ULC': int('114' in hcc_set) * diagnostic_cats['PRESSURE_ULCER'],
            'SEPSIS_ASP_SPEC_BACT_PNEUM': diagnostic_cats['SEPSIS'] * int('114' in hcc_set),
            'SCHIZOPHRENIA_COPD': int('57' in hcc_set) * diagnostic_cats['COPD'],
            'SCHIZOPHRENIA_CHF': int('57' in hcc_set) * diagnostic_cats['CHF'],
            'SCHIZOPHRENIA_SEIZURES': int('57' in hcc_set) * int('79' in hcc_set),
            'NONAGED_HCC85': demographics.non_aged * int('85' in hcc_set),
            'NONAGED_PRESSURE_ULCER': demographics.non_aged * diagnostic_cats['PRESSURE_ULCER'],
            'NONAGED_HCC161': demographics.non_aged * int('161' in hcc_set),
            'NONAGED_HCC39': demographics.non_aged * int('39' in hcc_set),
            'NONAGED_HCC77': demographics.non_aged * int('77' in hcc_set)
        })
            
    elif model_name == "RxHCC Model V08":
        # RxHCC NonAged interactions
        interactions.update({
            'NonAged_RXHCC1': demographics.non_aged * int('1' in hcc_set),
            'NonAged_RXHCC130': demographics.non_aged * int('130' in hcc_set),
            'NonAged_RXHCC131': demographics.non_aged * int('131' in hcc_set),
            'NonAged_RXHCC132': demographics.non_aged * int('132' in hcc_set),
            'NonAged_RXHCC133': demographics.non_aged * int('133' in hcc_set),
            'NonAged_RXHCC159': demographics.non_aged * int('159' in hcc_set),
            'NonAged_RXHCC163': demographics.non_aged * int('163' in hcc_set)
        })
    
    return interactions

def apply_interactions(demographics: Demographics, 
                      hcc_set: set[str], 
                      model_name: ModelName = "CMS-HCC Model V28") -> dict:
    """
    Calculate HCC interactions across CMS models. Handles CMS-HCC, ESRD, and RxHCC models.
    """
    # Start with demographic interactions
    interactions = create_demographic_interactions(demographics)
    
    # Add dual status interactions
    interactions.update(create_dual_interactions(demographics))
    
    # Get diagnostic categories for the model
    diagnostic_cats = get_diagnostic_categories(model_name, hcc_set)
    
    interactions.update(create_disease_interactions(model_name, diagnostic_cats, demographics, hcc_set))
        
    # Add HCC counts
    interactions.update(create_hcc_counts(hcc_set))
    
    return interactions


from typing import List, Set
from hccinfhir.datamodels import ServiceLevelData
from hccinfhir.utils import load_proc_filtering

# use import importlib.resources to load the professional_cpt_fn file as a list of strings
professional_cpt_default_fn = 'ra_eligible_cpt_hcpcs_2025.csv'
professional_cpt_default = load_proc_filtering(professional_cpt_default_fn)

def apply_filter(
    data: List[ServiceLevelData], 
    inpatient_tob: Set[str] = {'11X', '41X'},
    outpatient_tob: Set[str] = {'12X', '13X', '43X', '71X', '73X', '76X', '77X', '85X', '87X'},
    professional_cpt: Set[str] = professional_cpt_default
) -> List[ServiceLevelData]:
    # tob (Type of Bill) Filter is based on:
    # https://www.hhs.gov/guidance/sites/default/files/hhs-guidance-documents/2012181486-wq-092916_ra_webinar_slides_5cr_092816.pdf
    # https://www.hhs.gov/guidance/sites/default/files/hhs-guidance-documents/FinalEncounterDataDiagnosisFilteringLogic.pdf
    # https://www.cms.gov/files/document/encounterdatasystemedit20495andedit01415andtob87x07162021.pdf for 87X
    # NOTE: If no facility_type or service_type, then the claim is professional, in our implementation.
    # NOTE: The original CMS logic is for the "record" level, not the service level.
    #  Thus, when preparing the service level data, put all diagnosis codes into the diagnosis field.

    filtered_data = []
    for item in data:
        item_tob = '?' if item.facility_type is None else item.facility_type 
        item_tob += '?' if item.service_type is None else item.service_type 
        item_tob += 'X'

        if '?' in item_tob: # professional claims
            if item.procedure_code in professional_cpt:
                filtered_data.append(item)
        else:
            if item_tob in inpatient_tob:
                filtered_data.append(item)
            elif item_tob in outpatient_tob:
                if item.procedure_code in professional_cpt:
                    filtered_data.append(item)
            
    return filtered_data


from typing import Union, List, Literal
from hccinfhir.datamodels import ServiceLevelData
from hccinfhir.extractor_837 import extract_sld_837
from hccinfhir.extractor_fhir import extract_sld_fhir

def extract_sld(
    data: Union[str, dict], 
    format: Literal["837", "fhir"] = "fhir"
) -> List[ServiceLevelData]:
    """
    Unified entry point for SLD extraction with explicit format specification
    
    Args:
        data: Input data - string for 837, dict for FHIR
        format: Data format - either "837" or "fhir"
        
    Returns:
        List of ServiceLevelData
        
    Raises:
        ValueError: If format and data type don't match or format is invalid
        TypeError: If data is None or wrong type
    """
    if data is None:
        raise TypeError("Input data cannot be None")
        
    if format == "837":
        if not isinstance(data, str) or data == "":
            raise TypeError(f"837 format requires string input, got {type(data)}")
        return extract_sld_837(data)
    elif format == "fhir":
        if not isinstance(data, dict) or data == {}:
            raise TypeError(f"FHIR format requires dict input, got {type(data)}")   
        return extract_sld_fhir(data)
    else:
        raise ValueError(f'Format must be either "837" or "fhir", got {format}')


def extract_sld_list(data: Union[List[str], List[dict]], 
                     format: Literal["837", "fhir"] = "fhir") -> List[ServiceLevelData]:
    """Extract SLDs from a list of FHIR EOBs"""
    output = []
    for item in data:
        try:
            output.extend(extract_sld(item, format))
        except TypeError as e:
            print(f"Warning: Skipping invalid types: {str(e)}")
        except ValueError as e:
            print(f"Warning: Skipping invalid values: {str(e)}")
    return output




from typing import Dict, Tuple
import importlib.resources
from hccinfhir.datamodels import ModelName, Demographics

# Load default mappings from csv file
coefficients_file_default = 'ra_coefficients_2026.csv'
coefficients_default: Dict[Tuple[str, ModelName], float] = {}  # (diagnosis_code, model_name) -> value

try:
    with importlib.resources.open_text('hccinfhir.data', coefficients_file_default) as f:
        for line in f.readlines()[1:]:  # Skip header
            try:
                coefficient, value, model_domain, model_version = line.strip().split(',')
                if model_domain == 'ESRD':  
                    model_name = f"CMS-HCC {model_domain} Model V{model_version[-2:]}"
                else:
                    model_name = f"{model_domain} Model V{model_version[-2:]}"
                
                key = (coefficient.lower(), model_name)
                if key not in coefficients_default:
                    coefficients_default[key] = float(value)
                else:
                    coefficients_default[key] = float(value)
            except ValueError:
                continue  # Skip malformed lines
except Exception as e:
    print(f"Error loading mapping file: {e}")
    coefficients_default = {}

def get_coefficent_prefix(demographics: Demographics, 
                          model_name: ModelName = "CMS-HCC Model V28") -> str:

    """
    Get the coefficient prefix based on beneficiary demographics.
    
    Args:
        demographics: Demographics object containing beneficiary information
        
    Returns:
        String prefix used to look up coefficients for this beneficiary type
    """
    # Get base prefix based on model type
    if 'ESRD' in model_name:
        if demographics.esrd:
            if demographics.graft_months is not None:
                # Functioning graft case
                if demographics.lti:
                    return 'GI_'
                if demographics.new_enrollee:
                    return 'GNE_'
                    
                # Community functioning graft
                prefix = 'G'
                prefix += 'F' if demographics.fbd else 'NP'
                prefix += 'A' if demographics.age >= 65 else 'N'
                return prefix + '_'
                
            # Dialysis case
            return 'DNE_' if demographics.new_enrollee else 'DI_'
            
        # Transplant case
        if demographics.graft_months in [1, 2, 3]:
            return f'TRANSPLANT_KIDNEY_ONLY_{demographics.graft_months}M'
            
    elif 'RxHCC' in model_name:
        if demographics.lti:
            return 'Rx_NE_LTI_' if demographics.new_enrollee else 'Rx_CE_LTI_'
            
        if demographics.new_enrollee:
            return 'Rx_NE_Lo_' if demographics.low_income else 'Rx_NE_NoLo_'
            
        # Community case
        prefix = 'Rx_CE_'
        prefix += 'Low' if demographics.low_income else 'NoLow'
        prefix += 'Aged' if demographics.age >= 65 else 'NoAged'
        return prefix + '_'
        
    # Default CMS-HCC Model
    if demographics.lti:
        return 'INS_'
        
    if demographics.new_enrollee:

        return 'SNPNE_' if demographics.snp else 'NE_'
        
    # Community case
    prefix = 'C'
    prefix += 'F' if demographics.fbd else ('P' if demographics.pbd else 'N')
    prefix += 'A' if demographics.age >= 65 else 'D'
    return prefix + '_'


def apply_coefficients(demographics: Demographics, 
                      hcc_set: set[str], 
                      interactions: dict,
                      model_name: ModelName = "CMS-HCC Model V28",
                      coefficients: Dict[Tuple[str, ModelName], float] = coefficients_default) -> dict:
    """Apply risk adjustment coefficients to HCCs and interactions.

    This function takes demographic information, HCC codes, and interaction variables and returns
    a dictionary mapping each variable to its corresponding coefficient value based on the 
    specified model.

    Args:
        demographics: Demographics object containing patient characteristics
        hcc_set: Set of HCC codes present for the patient
        interactions: Dictionary of interaction variables and their values (0 or 1)
        model_name: Name of the risk adjustment model to use (default: "CMS-HCC Model V28")
        coefficients: Dictionary mapping (variable, model) tuples to coefficient values
            (default: coefficients_default)

    Returns:
        Dictionary mapping HCC codes and interaction variables to their coefficient values
        for variables that are present (HCC in hcc_set or interaction value = 1)
    """
    # Get the coefficient prefix
    prefix = get_coefficent_prefix(demographics, model_name)
    
    output = {}

    demographics_key = (f"{prefix}{demographics.category}".lower(), model_name)
    if demographics_key in coefficients:
        output[demographics.category] = coefficients[demographics_key]

    # Apply the coefficients
    for hcc in hcc_set:
        key = (f"{prefix}HCC{hcc}".lower(), model_name)

        if key in coefficients:
            value = coefficients[key]
            output[hcc] = value

    # Add interactions
    for interaction_key, interaction_value in interactions.items():
        if interaction_value < 1:
            continue

        key = (f"{prefix}{interaction_key}".lower(), model_name)
        if key in coefficients:
            value = coefficients[key]
            output[interaction_key] = value


    return output




from typing import Union
from hccinfhir.datamodels import Demographics
    
def categorize_demographics(age: Union[int, float], 
                       sex: str, 
                       dual_elgbl_cd: str = None,
                       orec: str = None, 
                       crec: str = None,
                       version: str = 'V2',
                       new_enrollee: bool = False,
                       snp: bool = False,
                       low_income: bool = False,
                       graft_months: int = None
                       ) -> Demographics:
    """
    Categorize a beneficiary's demographics into risk adjustment categories.

    This function takes demographic information about a beneficiary and returns a Demographics
    object containing derived fields used in risk adjustment models.

    Args:
        age: Beneficiary age (integer or float, will be floored to integer)
        sex: Beneficiary sex ('M'/'F' or '1'/'2')
        dual_elgbl_cd: Dual eligibility code ('00'-'10')
        orec: Original reason for entitlement code ('0'-'3')
        crec: Current reason for entitlement code ('0'-'3') 
        version: Version of categorization to use ('V2', 'V4', 'V6')
        new_enrollee: Whether beneficiary is a new enrollee
        snp: Whether beneficiary is in a Special Needs Plan

    Returns:
        Demographics object containing derived fields like age/sex category,
        disability status, dual status flags, etc.

    Raises:
        ValueError: If age is negative or non-numeric, or if sex is invalid
    """
    
    if not isinstance(age, (int, float)):
        raise ValueError("Age must be a number")
    
    if age < 0:
        raise ValueError("Age must be non-negative")
        
    # Convert to integer using floor
    age = int(age)
    non_aged = age <= 64

    # Standardize sex input
    if sex in ('M', '1'):
        std_sex = '1'  # For V2/V4
        v6_sex = 'M'   # For V6
    elif sex in ('F', '2'):
        std_sex = '2'  # For V2/V4
        v6_sex = 'F'   # For V6
    else:
        raise ValueError("Sex must be 'M', 'F', '1', or '2'")
    
    # Determine if person is disabled or originally disabled
    # SAS code:
    # DISABL = (&AGEF < 65 & &OREC ne "0");
    # ORIGDS  = (&OREC = '1')*(DISABL = 0);
    # The vairable names can be misleading. 
    # disabled is true if the person is disabled and the age is less than 65
    # - basically, the person is in Medicare due to disability not due to age
    # orig_disabled is true if the person started Medicare due to disability, but now aged in
    # - basically, the person is in Medicare due to age (not disability anymore)
    disabled = age < 65 and (orec is not None and orec != "0")
    orig_disabled = (orec is not None and orec == '1') and not disabled

    # Reference: https://resdac.org/cms-data/variables/medicare-medicaid-dual-eligibility-code-january 
    # Full benefit dual codes
    fbd_codes = {'02', '04', '08'}
    
    # Partial benefit dual codes
    pbd_codes = {'01', '03', '05', '06'}
    
    is_fbd = dual_elgbl_cd in fbd_codes
    is_pbd = dual_elgbl_cd in pbd_codes

    esrd_orec = orec in {'2', '3', '6'}
    esrd_crec = crec in {'2', '3'} if crec else False
    esrd = esrd_orec or esrd_crec

    result_dict = {
        'version': version,
        'non_aged': non_aged,
        'orig_disabled': orig_disabled,
        'disabled': disabled,
        'age': age,
        'sex': std_sex if version in ('V2', 'V4') else v6_sex,
        'dual_elgbl_cd': dual_elgbl_cd,
        'orec': orec,
        'crec': crec,
        'new_enrollee': new_enrollee,
        'snp': snp,
        'fbd': is_fbd,
        'pbd': is_pbd,
        'esrd': esrd,
        'graft_months': graft_months,
        'low_income': low_income
    }

    # V6 Logic (ACA Population)
    if version == 'V6':
        age_ranges = [
            (0, 0, '0_0'),
            (1, 1, '1_1'),
            (2, 4, '2_4'),
            (5, 9, '5_9'),
            (10, 14, '10_14'),
            (15, 20, '15_20'),
            (21, 24, '21_24'),
            (25, 29, '25_29'),
            (30, 34, '30_34'),
            (35, 39, '35_39'),
            (40, 44, '40_44'),
            (45, 49, '45_49'),
            (50, 54, '50_54'),
            (55, 59, '55_59'),
            (60, float('inf'), '60_GT')
        ]
        
        for low, high, label in age_ranges:
            if low <= age <= high:
                result_dict['category'] = f"{v6_sex}AGE_LAST_{label}"
                return Demographics(**result_dict)
    
    # V2/V4 Logic (Medicare Population)
    elif version in ('V2', 'V4'):
        if orec is None or orec == '':
            orec = '0' # Default to 0 if OREC is None

        # New enrollee logic
        if new_enrollee:
            prefix = 'NEF' if std_sex == '2' else 'NEM'
            
            if age <= 34:
                category = f'{prefix}0_34'
            elif 34 < age <= 44:
                category = f'{prefix}35_44'
            elif 44 < age <= 54:
                category = f'{prefix}45_54'
            elif 54 < age <= 59:
                category = f'{prefix}55_59'
            elif (59 < age <= 63) or (age == 64 and orec != '0'):
                category = f'{prefix}60_64'
            elif (age == 64 and orec == '0') or age == 65:
                category = f'{prefix}65'
            elif age == 66:
                category = f'{prefix}66'
            elif age == 67:
                category = f'{prefix}67'
            elif age == 68:
                category = f'{prefix}68'
            elif age == 69:
                category = f'{prefix}69'
            elif 69 < age <= 74:
                category = f'{prefix}70_74'
            elif 74 < age <= 79:
                category = f'{prefix}75_79'
            elif 79 < age <= 84:
                category = f'{prefix}80_84'
            elif 84 < age <= 89:
                category = f'{prefix}85_89'
            elif 89 < age <= 94:
                category = f'{prefix}90_94'
            else:
                category = f'{prefix}95_GT'
        
        else:
            prefix = 'F' if std_sex == '2' else 'M'
            age_ranges = [
                (0, 34, '0_34'),
                (34, 44, '35_44'),
                (44, 54, '45_54'),
                (54, 59, '55_59'),
                (59, 64, '60_64'),
                (64, 69, '65_69'),
                (69, 74, '70_74'),
                (74, 79, '75_79'),
                (79, 84, '80_84'),
                (84, 89, '85_89'),
                (89, 94, '90_94'),
                (94, float('inf'), '95_GT')
            ]
            
            for low, high, suffix in age_ranges:
                if low < age <= high:
                    category = f'{prefix}{suffix}'
                    break
            else:
                raise ValueError(f"Unable to categorize age: {age}")
        
        result_dict['category'] = category
        return Demographics(**result_dict)
    
    else:
        raise ValueError("Version must be 'V2', 'V4', or 'V6'")


from typing import List, Union, Dict, Tuple, Set
from hccinfhir.datamodels import ModelName, RAFResult
from hccinfhir.model_demographics import categorize_demographics
from hccinfhir.model_dx_to_cc import apply_mapping
from hccinfhir.model_hierarchies import apply_hierarchies
from hccinfhir.model_coefficients import apply_coefficients
from hccinfhir.model_interactions import apply_interactions
from hccinfhir.utils import load_dx_to_cc_mapping, load_is_chronic

# Load default mappings from csv file
mapping_file_default = 'ra_dx_to_cc_2026.csv'
dx_to_cc_default = load_dx_to_cc_mapping(mapping_file_default)

# Load default mappings from csv file
mapping_file_default = 'hcc_is_chronic.csv'
is_chronic_default = load_is_chronic(mapping_file_default)

def calculate_raf(diagnosis_codes: List[str],
                  model_name: ModelName = "CMS-HCC Model V28",
                  age: Union[int, float] = 65, 
                  sex: str = 'F', 
                  dual_elgbl_cd: str = 'NA',
                  orec: str = '0', 
                  crec: str = '0',
                  new_enrollee: bool = False,   
                  snp: bool = False,
                  low_income: bool = False,
                  graft_months: int =  None,
                  dx_to_cc_mapping: Dict[Tuple[str, ModelName], Set[str]] = dx_to_cc_default,
                  is_chronic_mapping: Dict[Tuple[str, ModelName], bool] = is_chronic_default) -> RAFResult:
    """
    Calculate Risk Adjustment Factor (RAF) based on diagnosis codes and demographic information.

    Args:
        diagnosis_codes: List of ICD-10 diagnosis codes
        model_name: Name of the HCC model to use
        age: Patient's age
        sex: Patient's sex ('M' or 'F')
        dual_elgbl_cd: Dual eligibility code
        orec: Original reason for entitlement code
        crec: Current reason for entitlement code
        new_enrollee: Whether the patient is a new enrollee
        snp: Special Needs Plan indicator
        low_income: Low income subsidy indicator
        graft_months: Number of months since transplant

    Returns:
        Dictionary containing RAF score and coefficients used in calculation

    Raises:
        ValueError: If input parameters are invalid
    """
    # Input validation
    if not isinstance(age, (int, float)) or age < 0:
        raise ValueError("Age must be a non-negative number")
    
    if sex not in ['M', 'F', '1', '2']:
        raise ValueError("Sex must be 'M' or 'F' or '1' or '2'")

    version = 'V2'
    if 'RxHCC' in model_name:
        version = 'V4'
    elif 'HHS-HCC' in model_name: # not implemented yet
        version = 'V6'
    
    demographics = categorize_demographics(age, 
                                           sex, 
                                           dual_elgbl_cd, 
                                           orec, 
                                           crec, 
                                           version, 
                                           new_enrollee, 
                                           snp, 
                                           low_income, 
                                           graft_months)
    
    cc_to_dx = apply_mapping(diagnosis_codes, 
                             model_name, 
                             dx_to_cc_mapping=dx_to_cc_mapping)
    hcc_set = set(cc_to_dx.keys())
    hcc_set = apply_hierarchies(hcc_set, model_name)
    interactions = apply_interactions(demographics, hcc_set, model_name)
    coefficients = apply_coefficients(demographics, hcc_set, interactions, model_name)

    hcc_chronic = set()
    for hcc in hcc_set:
        if is_chronic_mapping.get((hcc, model_name), False):
            hcc_chronic.add(hcc)

    demographic_interactions = {}
    for key, value in interactions.items():
        if key.startswith('NMCAID_'):
            demographic_interactions[key] = value
        elif key.startswith('MCAID_'):
            demographic_interactions[key] = value
        elif key.startswith('LTI_'):
            demographic_interactions[key] = value
        elif key.startswith('OriginallyDisabled_'):
            demographic_interactions[key] = value

    coefficients_demographics = apply_coefficients(demographics, 
                                                   set(), 
                                                   demographic_interactions, 
                                                   model_name)
    coefficients_chronic_only = apply_coefficients(demographics, 
                                                   hcc_chronic, 
                                                   demographic_interactions, 
                                                   model_name)
    
    # Calculate risk scores
    risk_score = sum(coefficients.values())
    risk_score_demographics = sum(coefficients_demographics.values())
    risk_score_chronic_only = sum(coefficients_chronic_only.values()) - risk_score_demographics
    risk_score_hcc = risk_score - risk_score_demographics

    return {
        'risk_score': risk_score, 
        'risk_score_demographics': risk_score_demographics,
        'risk_score_chronic_only': risk_score_chronic_only,
        'risk_score_hcc': risk_score_hcc,
        'hcc_list': list(hcc_set),
        'cc_to_dx': cc_to_dx,
        'coefficients': coefficients,
        'interactions': interactions,
        'demographics': demographics,
        'model_name': model_name,
        'version': version,
        'diagnosis_codes': diagnosis_codes,
    }






from typing import List, Dict, Any, Union
from hccinfhir.extractor import extract_sld_list
from hccinfhir.filter import apply_filter
from hccinfhir.model_calculate import calculate_raf
from hccinfhir.datamodels import Demographics, ServiceLevelData, RAFResult, ModelName, ProcFilteringFilename, DxCCMappingFilename
from hccinfhir.utils import load_proc_filtering, load_dx_to_cc_mapping

class HCCInFHIR:
    """
    Main class for processing FHIR EOB resources into HCC risk scores.
    
    This class integrates the extraction, filtering, and calculation components
    of the hccinfhir library.
    """
    
    def __init__(self, 
                 filter_claims: bool = True, 
                 model_name: ModelName = "CMS-HCC Model V28",
                 proc_filtering_filename: ProcFilteringFilename = "ra_eligible_cpt_hcpcs_2026.csv",
                 dx_cc_mapping_filename: DxCCMappingFilename = "ra_dx_to_cc_2026.csv"):
        """
        Initialize the HCCInFHIR processor.
        
        Args:
            filter_claims: Whether to apply filtering rules to claims. Default is True.
            model_name: The name of the model to use for the calculation. Default is "CMS-HCC Model V28".
            proc_filtering_filename: The filename of the professional cpt filtering file. Default is "ra_eligible_cpt_hcpcs_2026.csv".
            dx_cc_mapping_filename: The filename of the dx to cc mapping file. Default is "ra_dx_to_cc_2026.csv".
        """
        self.filter_claims = filter_claims
        self.model_name = model_name
        self.proc_filtering_filename = proc_filtering_filename
        self.dx_cc_mapping_filename = dx_cc_mapping_filename
        self.professional_cpt = load_proc_filtering(proc_filtering_filename)
        self.dx_to_cc_mapping = load_dx_to_cc_mapping(dx_cc_mapping_filename)


    def _ensure_demographics(self, demographics: Union[Demographics, Dict[str, Any]]) -> Demographics:
        """Convert demographics dict to Demographics object if needed."""
        if not isinstance(demographics, Demographics):
            return Demographics(**demographics)
        return demographics
    
    def _calculate_raf_from_demographics(self, diagnosis_codes: List[str], 
                                       demographics: Demographics) -> RAFResult:
        """Calculate RAF score using demographics data."""
        return calculate_raf(
            diagnosis_codes=diagnosis_codes,
            model_name=self.model_name,
            age=demographics.age,
            sex=demographics.sex,
            dual_elgbl_cd=demographics.dual_elgbl_cd,
            orec=demographics.orec,
            crec=demographics.crec,
            new_enrollee=demographics.new_enrollee,
            snp=demographics.snp,
            low_income=demographics.low_income,
            graft_months=demographics.graft_months,
            dx_to_cc_mapping=self.dx_to_cc_mapping
        )

    def _get_unique_diagnosis_codes(self, service_data: List[ServiceLevelData]) -> List[str]:
        """Extract unique diagnosis codes from service level data."""
        return list({code for sld in service_data for code in sld.claim_diagnosis_codes})

    def run(self, eob_list: List[Dict[str, Any]], 
            demographics: Union[Demographics, Dict[str, Any]]) -> RAFResult:
        """Process EOB resources and calculate RAF scores.
        
        Args:
            eob_list: List of EOB resources
            demographics: Demographics information
            
        Returns:
            RAFResult object containing calculated scores and processed data
        """
        if not isinstance(eob_list, list):
            raise ValueError("eob_list must be a list; if no eob, pass empty list")
        
        demographics = self._ensure_demographics(demographics)
        
        # Extract and filter service level data
        sld_list = extract_sld_list(eob_list)

        if self.filter_claims:
            sld_list = apply_filter(sld_list, professional_cpt=self.professional_cpt)
            
        # Calculate RAF score
        unique_dx_codes = self._get_unique_diagnosis_codes(sld_list)
        raf_result = self._calculate_raf_from_demographics(unique_dx_codes, demographics)
        raf_result['service_level_data'] = sld_list
        return raf_result
    
    def run_from_service_data(self, service_data: List[Union[ServiceLevelData, Dict[str, Any]]], 
                             demographics: Union[Demographics, Dict[str, Any]]) -> RAFResult:
        demographics = self._ensure_demographics(demographics)
        
        if not isinstance(service_data, list):
            raise ValueError("Service data must be a list of service records")
                
        # Standardize service data with better error handling
        standardized_data = []
        for idx, item in enumerate(service_data):
            try:
                if isinstance(item, dict):
                    standardized_data.append(ServiceLevelData(**item))
                elif isinstance(item, ServiceLevelData):
                    standardized_data.append(item)
                else:
                    raise TypeError(f"Service data item must be a dictionary or ServiceLevelData object")
            except (KeyError, TypeError, ValueError) as e:
                raise ValueError(
                    f"Invalid service data at index {idx}: {str(e)}. "
                    "Required fields: claim_type, claim_diagnosis_codes, procedure_code, service_date"
                )
        
        if self.filter_claims:
            standardized_data = apply_filter(standardized_data, 
                                             professional_cpt=self.professional_cpt)

        
        # Calculate RAF score
        unique_dx_codes = self._get_unique_diagnosis_codes(standardized_data)
        raf_result = self._calculate_raf_from_demographics(unique_dx_codes, demographics)
        raf_result['service_level_data'] = standardized_data

        return raf_result
        
    def calculate_from_diagnosis(self, diagnosis_codes: List[str],
                               demographics: Union[Demographics, Dict[str, Any]]) -> RAFResult:
        """Calculate RAF scores from a list of diagnosis codes.
        
        Args:
            diagnosis_codes: List of diagnosis codes
            demographics: Demographics information
            
        Raises:
            ValueError: If diagnosis_codes is empty or not a list
        """
        if not isinstance(diagnosis_codes, list):
            raise ValueError("diagnosis_codes must be a list")
        if not diagnosis_codes:
            raise ValueError("diagnosis_codes list cannot be empty")
        
        demographics = self._ensure_demographics(demographics)
        raf_result = self._calculate_raf_from_demographics(diagnosis_codes, demographics)
        return raf_result